class com.cintegrity.Lib.Collection.List
    implements com.cintegrity.Lib.Collection.Collection
    :
  define private temp-table ctt_List
      field in_Sequence as integer
      field ob_AnObject as class Progress.Lang.Object
      index by_Sequence as primary unique in_Sequence
      index by_Object ob_AnObject.
  define private variable cin_CurrentSequence as integer no-undo initial 0.
  define private variable cin_HighestSequence as integer no-undo initial 0.

  destructor public List():
    empty temp-table ctt_List.
  end destructor.
  /* ---------------------------------------- implement Collection interface */
  method public integer size():
    return cin_HighestSequence.
  end method.

  method public logical isEmpty():
    return (cin_HighestSequence eq 0).
  end method.

  method public logical containsElement(
      input mob_Element as class Progress.Lang.Object
      ):
    find ctt_List where mob_Element eq ctt_List.ob_AnObject no-error.
    return available( ctt_List ).
  end method.

  method public logical addElement(
      input mob_Element as class Progress.Lang.Object
      ):
    create ctt_List.
    assign
        cin_HighestSequence = cin_HighestSequence + 1
        ctt_List.in_Sequence = cin_HighestSequence
        ctt_List.ob_AnObject = mob_Element
        cin_CurrentSequence = cin_HighestSequence
        .
    return true.
  end method.

  method public logical removeElement(
      input mob_Element as class Progress.Lang.Object
      ):
    define variable min_DeletedElementSeq as integer no-undo.
    find ctt_List where mob_Element eq ctt_List.ob_AnObject no-error.
    if available( ctt_List )
    then do:
      min_DeletedElementSeq = ctt_List.in_Sequence.
      delete ctt_List.
      if min_DeletedElementSeq LT cin_HighestSequence
      then do:
        for each ctt_List
            where ctt_List.in_Sequence GE min_DeletedElementSeq
            :
          ctt_List.in_Sequence = ctt_List.in_Sequence - 1.
        end.
        cin_HighestSequence = ctt_List.in_Sequence.
        cin_CurrentSequence = min_DeletedElementSeq + 1.
      end.
      else do:
        cin_HighestSequence = cin_HighestSequence - 1.
        cin_CurrentSequence = cin_HighestSequence.
      end.
      return true.
    end.
    else return false.
  end method.

  /* -------------------------------------------------- Implementation deferred
  method public logical containsAll(
      input mob_Collection as class com.cintegrity.Lib.Collection.Collection 
      ):
  end method.

  method public void addAll(
      input mob_Collection as class com.cintegrity.Lib.Collection.Collection
      ):
  end method.

  method public void removeAll(
      input mob_Collection as class com.cintegrity.Lib.Collection.Collection
      ):
  end method.

  method public void retainAll(
      input mob_Collection as class com.cintegrity.Lib.Collection.Collection
      ):
  end method.
  -------------------------------------------------------------------------- */

  method public void clearAll():
    empty temp-table ctt_List.
    cin_CurrentSequence = 0.
    cin_HighestSequence = 0.
  end method.

  method public logical hasNextElement():
    return can-find( first ctt_List where ctt_List.in_Sequence GT cin_CurrentSequence ).
  end method.

  method public logical hasPrevElement():
    return can-find( first ctt_List where ctt_List.in_Sequence LT cin_CurrentSequence ).
  end method.

  method public class Progress.Lang.Object nextElement( ):
    find ctt_List
        where ctt_List.in_Sequence = ( cin_CurrentSequence + 1 )
        no-error.
    if available( ctt_List )
    then do:
      cin_CurrentSequence = ctt_List.in_Sequence.
      return ctt_List.ob_AnObject.
    end.
    else do:
      /* raise "No Such Element" exception */
      return ?.
    end.
  end method.

  method public class Progress.Lang.Object prevElement( ):
    find ctt_List
        where ctt_List.in_Sequence = ( cin_CurrentSequence - 1 )
        no-error.
    if available( ctt_List )
    then do:
      cin_CurrentSequence = ctt_List.in_Sequence.
      return ctt_List.ob_AnObject.
    end.
    else do:
      /* raise "No Such Element" exception */
      return ?.
    end.
  end method.

  method public class Progress.Lang.Object setElement(
      input mob_Element as class Progress.Lang.Object
      ):
    define variable mob_OldObject as class Progress.Lang.Object no-undo.

    find ctt_List
        where ctt_List.in_Sequence = cin_CurrentSequence
        no-error.
    if available( ctt_List )
    then do:
      mob_OldObject = ctt_List.ob_AnObject.
      delete ctt_List.
      create ctt_List.
      assign
          ctt_List.in_Sequence = cin_CurrentSequence
          ctt_List.ob_AnObject = mob_Element
          .
      return mob_OldObject.
    end.
    else do:
      /* raise "No Current Element to Replace" exception */
    end.
  end method.
  /* ------------------------------------------- end of Collection interface */

  /* --------------------------------------------------- extensions for List */
  method public class Progress.Lang.Object getNthElement(
      input min_Index as integer
      ):
    cin_CurrentSequence = min_Index.
    find ctt_List
        where cin_CurrentSequence = ctt_List.in_Sequence
        no-error.
    if available( ctt_List )
    then return ctt_List.ob_AnObject.
    else do:
                                              /* raise No Such Element error */
      return ?.
    end.
  end method.

  method public class Progress.Lang.Object setNthElement(
      input min_Index as integer,
      input mob_Element as class Progress.Lang.Object
      ):
    define variable mob_OldObject as class Progress.Lang.Object no-undo.

    cin_CurrentSequence = min_Index.
    find ctt_List
        where cin_CurrentSequence = ctt_List.in_Sequence
        no-error.
    if available( ctt_List )
    then do:
      mob_OldObject = ctt_List.ob_AnObject.
      delete ctt_List.
      create ctt_List.
      assign
          ctt_List.in_Sequence = cin_CurrentSequence
          ctt_List.ob_AnObject = mob_Element
          .
      return mob_OldObject.
    end.
    else do:
      create ctt_List.
      assign
          ctt_List.in_Sequence = cin_CurrentSequence
          ctt_List.ob_AnObject = mob_Element
          .
      return ?.
    end.
  end method.

  method public void addNthElement(
      input min_Index as integer,
      input mob_Element as class Progress.Lang.Object
      ):
    define variable min_Element as integer no-undo.
    define buffer mbf_List for ctt_List.

    cin_CurrentSequence = min_Index.
    if can-find( first ctt_List where min_Index EQ ctt_List.in_Sequence  )
    then do min_Element = cin_HighestSequence to min_Index by -1:
      find mbf_List where mbf_List.in_Sequence eq min_Element no-error.
      if available( mbf_List )
      then do:
        mbf_List.in_Sequence = mbf_List.in_Sequence + 1.
      end.
    end.
    create ctt_List.
    assign
        ctt_List.in_Sequence = cin_CurrentSequence
        ctt_List.ob_AnObject = mob_Element
        .
  end method.

  method public Progress.Lang.Object removeNthElement(
      input min_Index as integer
      ):
    define variable mob_OldObject as class Progress.Lang.Object no-undo.
    define buffer mbf_List for ctt_List.

    cin_CurrentSequence = min_Index.
    find ctt_List
        where min_Index = ctt_List.in_Sequence
        no-error.
    if available( ctt_List )
    then do:
      mob_OldObject = ctt_List.ob_AnObject.
      delete ctt_List.
      for each mbf_List
          where mbf_List.in_Sequence ge min_Index
          by mbf_List.in_Sequence:
        mbf_List.in_Sequence = mbf_List.in_Sequence - 1.
      end.
      return mob_OldObject.
    end.
    else do:
                                              /* raise No Such Element error */
      return ?.
    end.
  end method.

  method public integer indexOf(
      input mob_Element as class Progress.Lang.Object
      ):
    for each ctt_List
        where ctt_List.ob_AnObject eq mob_Element
        by ctt_List.in_Sequence
        :
      leave.
    end.
    if available( ctt_List )
    then do:
      cin_CurrentSequence = ctt_List.in_Sequence.
      return cin_CurrentSequence.
    end.
    else do:
      /* raise No Such Element exception */
      return ?.
    end.
  end method.

  method public integer lastIndexOf(
      input mob_Element as class Progress.Lang.Object
      ):
    for each ctt_List
        where ctt_List.ob_AnObject eq mob_Element
        by ctt_List.in_Sequence descending
        :
      leave.
    end.
    if available( ctt_List )
    then do:
      cin_CurrentSequence = ctt_List.in_Sequence.
      return cin_CurrentSequence.
    end.
    else do:
      /* raise No Such Element exception */
      return ?.
    end.
  end method.

  /* -------------------------------------------------- Implementation deferred
  method public class com.cintegrity.Lib.Collection.List subList(
      input mob_MinIndex as integer,
      input mob_MaxIndex as integer
      ):
  end method.
  -------------------------------------------------------------------------- */

  method public override class Progress.Lang.Object clone():
    define variable mob_CloneList as com.cintegrity.Lib.Collection.List.
    define variable mlg_IsOk as logical no-undo.

    mob_CloneList = new com.cintegrity.Lib.Collection.List().
    for each ctt_List:
      mlg_IsOK = mob_CloneList:addElement( ctt_List.ob_AnObject ).
      if not mlg_IsOK
      then do:
                                           /* raise unsuccessful clone error */
        return ?.
      end.
    end.
    return mob_CloneList.

  end method.

end class.
